/*
  QtCurve (C) Craig Drummond, 2003   Craig.Drummond@lycos.co.uk

  ----

  Based upon B???Curve style (Copyright (c) 2002 R?d H?t, Inc)
      Bernhard Rosenkrnzer <bero@r?dh?t.com>
      Preston Brown <pbrown@r?dh?t.com>
      Than Ngo <than@r?dh?t.com>

  Released under the GNU General Public License (GPL) v2.

  ----

  B???Curve is based on the KDE Light style, 2nd revision:
  Copyright (c) 2000-2001 Trolltech AS (info@trolltech.com)

  Permission is hereby granted, free of charge, to any person obtaining a
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation
  the rights to use, copy, modify, merge, publish, distribute, sublicense,
  and/or sell copies of the Software, and to permit persons to whom the
  Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.
*/

#include <qmenubar.h>
#include <qapplication.h>
#include <qpainter.h>
#include <qpalette.h>
#include <qframe.h>
#include <qpushbutton.h>
#include <qdrawutil.h>
#include <qscrollbar.h>
#include <qtabbar.h>
#include <qguardedptr.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qlistview.h>
#include <qbitmap.h>
#include <qimage.h>
#include <qcombobox.h>
#include <qslider.h>
#include <qstylefactory.h>
#include <qcleanuphandler.h>
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <qtoolbutton.h>
#include <qtoolbar.h>
#include <qprogressbar.h>
#include "qtcurve.h"

#define CHECK_SIZE 13
#define RADIO_SIZE 12
#define HIGHLIGHT_FACTOR 105

#define MAX(x, y) ((x) > (y) ? (x) : (y))
#define MIN(x, y) ((x) < (y) ? (x) : (y))

static void rgb_to_hls (double *r, double *g, double *b)
{
  double min,
         max,
         red=*r,
         green=*g,
         blue=*b,
         h, l, s,
         delta;

  if (red > green)
    {
      if (red > blue)
        max = red;
      else
        max = blue;

      if (green < blue)
        min = green;
      else
        min = blue;
    }
  else
    {
      if (green > blue)
        max = green;
      else
        max = blue;

      if (red < blue)
        min = red;
      else
        min = blue;
    }

  l = (max + min) / 2;
  s = 0;
  h = 0;

  if (max != min)
    {
      if (l <= 0.5)
        s = (max - min) / (max + min);
      else
        s = (max - min) / (2 - max - min);

      delta = max -min;
      if (red == max)
        h = (green - blue) / delta;
      else if (green == max)
        h = 2 + (blue - red) / delta;
      else if (blue == max)
        h = 4 + (red - green) / delta;

      h *= 60;
      if (h < 0.0)
        h += 360;
    }

  *r = h;
  *g = l;
  *b = s;
}

static void hls_to_rgb (double *h, double *l, double *s)
{
  double hue,
         lightness=*l,
         saturation=*s,
         m1, m2,
         r, g, b;

  lightness = *l;
  saturation = *s;

  if (lightness <= 0.5)
    m2 = lightness * (1 + saturation);
  else
    m2 = lightness + saturation - lightness * saturation;
  m1 = 2 * lightness - m2;

  if (0==saturation)
  {
      *h = lightness;
      *l = lightness;
      *s = lightness;
  }
  else
  {
      hue = *h + 120;
      while (hue > 360)
        hue -= 360;
      while (hue < 0)
        hue += 360;

      if (hue < 60)
        r = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
        r = m2;
      else if (hue < 240)
        r = m1 + (m2 - m1) * (240 - hue) / 60;
      else
        r = m1;

      hue = *h;
      while (hue > 360)
        hue -= 360;
      while (hue < 0)
        hue += 360;

      if (hue < 60)
        g = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
        g = m2;
      else if (hue < 240)
        g = m1 + (m2 - m1) * (240 - hue) / 60;
      else
        g = m1;

      hue = *h - 120;
      while (hue > 360)
        hue -= 360;
      while (hue < 0)
        hue += 360;

      if (hue < 60)
        b = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
        b = m2;
      else if (hue < 240)
        b = m1 + (m2 - m1) * (240 - hue) / 60;
      else
        b = m1;

      *h = r;
      *l = g;
      *s = b;
  }
}

inline int limit(double c)
{
    return c < 0   ? 0   :
           c > 255 ? 255 :
           (int)c;
}

static void shade (const QColor &a, QColor *b, float k)
{
  double red=a.red()/256.0,
         green=a.green()/256.0,
         blue=a.blue()/256.0;

  rgb_to_hls (&red, &green, &blue);

  green *= k;
  if (green > 1.0)
    green = 1.0;
  else if (green < 0.0)
    green = 0.0;

  blue *= k;
  if (blue > 1.0)
    blue = 1.0;
  else if (blue < 0.0)
    blue = 0.0;

  hls_to_rgb (&red, &green, &blue);

  b->setRgb(limit(red*256), limit(green*256), limit(blue*256));

  // CPD... tried:
  //
  //int h, s, v;
  //
  //a.hsv(&h, &s, &v);
  //b.setHsv(h, limit((s*k)+0.5), limit((v*k)+0.5));
  //
  // ...but this gave slightly different values...
}

static void shadeColors(const QColor &base, QColor *vals)
{
  //double shades[] = {/*1.065*/ 1.1, 0.98, 0.896, 0.79, 0.768, 0.4 };
  double shades[] = {1.15, 1.05, 0.896, 0.79, 0.768, 0.4 };

  int i;

  for(i=0; i<6; ++i)
    shade (base, &vals[i], (shades[i]-0.7)*1.0+0.7);
  vals[6]=base;
}

static void shadeGradient(const QColor &base, QColor *vals)
{
    vals[QtCurveStyle::GRADIENT_BASE]=base;
    shade(vals[QtCurveStyle::GRADIENT_BASE], &(vals[QtCurveStyle::GRADIENT_TOP]), 0.92);
    shade(vals[QtCurveStyle::GRADIENT_BASE], &(vals[QtCurveStyle::GRADIENT_BOTTOM]), 1.66);
    shade(vals[QtCurveStyle::GRADIENT_BASE], &(vals[QtCurveStyle::GRADIENT_LIGHT]), 1.62);
    shade(vals[QtCurveStyle::GRADIENT_BASE], &(vals[QtCurveStyle::GRADIENT_DARK]), 1.05);
}

static void drawArrow(QPainter *p, const QRect &r, const QColor &col, QStyle::PrimitiveElement pe, bool small=false)
{
    QPointArray a;

    if(small)
        switch(pe)
        {
            case QStyle::PE_ArrowUp:
                a.setPoints(7,  2,1,  2,0,  0,-2,  -2,0,  -2,1,  -2,0,  2,0);
                break;
            case QStyle::PE_ArrowDown:
                a.setPoints(7,  2,-1,  2,0,  0,2,  -2,0,  -2,-1,  -2,0,  2,0);
                break;
            case QStyle::PE_ArrowRight:
                a.setPoints(7,  1,-2,  0,-2,  -2,0,  0,2,  1,2,  0,2,  0,-2);
                break;
            case QStyle::PE_ArrowLeft:
                a.setPoints(7,  -1,-2,  0,-2,  2,0,  0,2,  -1,2,  0,2,  0,-2);
                break;
            default:
                return;
        }
    else
        switch(pe)
        {
            case QStyle::PE_ArrowUp:
                a.setPoints(7,  3,1,  0,-2,  -3,1,  -2,2,  -1,1,  1,1,  2,2);
                break;
            case QStyle::PE_ArrowDown:
                a.setPoints(7,  3,-1,  0,2,  -3,-1,  -2,-2,  -1,-1,  1,-1,  2,-2);
                break;
            case QStyle::PE_ArrowRight:  
                a.setPoints(7,  -1,-3,  2,0,  -1,3,  -2,2,  -1,1,  -1,-1,  -2,-2);
                break;
            case QStyle::PE_ArrowLeft:
                a.setPoints(7,  1,-3,  -2,0,  1,3,  2,2,  1,1,  1,-1,  2,-2);
                break;
            default:
                return;
        }

    if (a.isNull())
        return;

    p->save();
    a.translate((r.x() + r.width() / 2), (r.y() + r.height() / 2));
    p->setBrush(col);
    p->setPen(col);
    p->drawPolygon(a);
    p->restore();
}

static void drawLines(QPainter *p, const QRect &r, bool horiz, int nLines, int offset, const QColor *cols, int startOffset=0)
{
    int space = (nLines*2)+(nLines-1),
        x = horiz ? r.x() : r.x()+((r.width()-space)/2),
        y = horiz ? r.y()+((r.height()-space)/2) : r.y(),
        x2 = r.x() + r.width()-1,
        y2 = r.y() + r.height()-1,
        i;

    if (horiz)
    {
        if(startOffset && y+startOffset>0)
            y+=startOffset;

        p->setPen(cols[0]);
        for(i=0; i<space; i+=3)
            p->drawLine(x+offset, y+i, x2-offset, y+i);

        p->setPen(cols[4]);
        for(i=1; i<space; i+=3)
            p->drawLine(x+offset, y+i, x2-offset, y+i);
    }
    else
    {
        if(startOffset && x+startOffset>0)
            x+=startOffset;

        p->setPen(cols[0]);
        for(i=0; i<space; i+=3)
            p->drawLine(x+i, y+offset, x+i, y2-offset);

        p->setPen(cols[4]);
        for(i=1; i<space; i+=3)
            p->drawLine(x+i, y+offset, x+i, y2-offset);
    }
}

QtCurveStyle::QtCurveStyle(bool bb, bool bf)
            : KStyle(AllowMenuTransparency, WindowsStyleScrollBar),
              isKicker(0==qstrcmp(qApp->argv()[0], "kicker") || 0==qstrcmp(qApp->argv()[0], "appletproxy")),
              borderButton(bb), borderFrame(bf),
              hoverWidget(NULL)
{
    shadeGradient(qApp->palette().active().highlight(), gradient);
    shadeColors(qApp->palette().active().background(), gray);
}

void QtCurveStyle::polish(QPalette &)
{
    if(qApp->palette().active().highlight()!=gradient[GRADIENT_BASE])
        shadeGradient(qApp->palette().active().highlight(), gradient);

    if(gray[6]!=qApp->palette().active().background())
        shadeColors(qApp->palette().active().background(), gray);

    if(button[6]!=qApp->palette().active().button())
        shadeColors(qApp->palette().active().button(), button);
}

static const char* kdeToolbarWidget = "kde toolbar widget";

void QtCurveStyle::polish(QWidget *widget)
{
    if (widget->inherits("QButton") || widget->inherits("QComboBox"))
        widget->installEventFilter(this);
    else if (widget->inherits("QMenuBar") || widget->inherits("QToolBar"))
        widget->setBackgroundMode(QWidget::PaletteBackground);
    else if (widget->inherits("QPopupMenu"))
        widget->setBackgroundMode(QWidget::NoBackground);
    else if (widget->inherits("QScrollBar") || widget->inherits("QSlider"))
    {
        widget->setMouseTracking(true);
        widget->installEventFilter(this);
    }
    else if (0==qstrcmp( widget->name(), kdeToolbarWidget))
        widget->setBackgroundMode(PaletteBackground);  // We paint the whole background.

    KStyle::polish(widget);
}

void QtCurveStyle::unPolish(QWidget *widget)
{
    if (widget->inherits("QButton") || widget->inherits("QComboBox"))
        widget->removeEventFilter(this);
    else if (widget->inherits("QMenuBar") || widget->inherits("QToolBar"))
        widget->setBackgroundMode(QWidget::PaletteButton);
    else if (widget->inherits("QPopupMenu"))
        widget->setBackgroundMode(QWidget::PaletteBackground);
    else if (widget->inherits("QScrollBar") || widget->inherits("QSlider"))
    {
        widget->setMouseTracking(false);
        widget->removeEventFilter(this);
    }
    else if (0==qstrcmp( widget->name(), kdeToolbarWidget))
    {
        widget->removeEventFilter(this);
        widget->setBackgroundMode( PaletteBackground );
    }
    KStyle::unPolish(widget);
}

bool QtCurveStyle::eventFilter(QObject *object, QEvent *event)
{
    switch(event->type())
    {
        case QEvent::Enter:
            if (object->isWidgetType())
            {
                hoverWidget=(QWidget *)object;
                if (hoverWidget->isEnabled())
                    hoverWidget->repaint(false);
                else
                    hoverWidget = NULL;
            }
            break;
        case QEvent::Leave:
            if (hoverWidget && object == hoverWidget)
            {
                hoverWidget = NULL;
                ((QWidget *)object)->repaint(false);
            }
            break;
        default:
            break;
    }
    return KStyle::eventFilter(object, event);
}

void QtCurveStyle::drawLightBevel(QPainter *p, const QRect &r,
                                  QStyle::SFlags flags, const QColor *fill, const QColor *custom) const
{
    QRect br(r);
    bool  sunken = (flags & (QStyle::Style_Down | QStyle::Style_On | QStyle::Style_Sunken));
    int   dark = borderButton ? 4 : 5;

    p->save();

    if(borderButton)
    {
        p->setPen(custom ? custom[5] : gray[5]);
        p->drawRect(r);
    }
    else
        br.addCoords(-1, -1, 1, 1);

    if ((sunken && !borderButton) || (!sunken && flags & QStyle::Style_Raised))
    {
        int c1=sunken ? dark : 0,
            c2=sunken ? 0 : dark;

	// button bevel
	p->setPen(custom ? custom[c1] : gray[c1]);
	p->drawLine(br.x() + 1, br.y() + 2,
		    br.x() + 1, br.y() + br.height() - 3); // left
	p->drawLine(br.x() + 1, br.y() + 1,
		    br.x() + br.width() - 2, br.y() + 1); // top

	p->setPen(custom ? custom[c2] : gray[c2]);
	p->drawLine(br.x() + br.width() - 2, br.y() + 1,
		    br.x() + br.width() - 2, br.y() + br.height() - 3); // right
	p->drawLine(br.x() + 1, br.y() + br.height() - 2,
		    br.x() + br.width() - 2, br.y() + br.height() - 2); // bottom

	br.addCoords(2, 2, -2, -2);
    }
    else
	br.addCoords(1, 1, -1, -1);

    // fill
    if (fill)
        p->fillRect(br, *fill);

    p->restore();
}

void QtCurveStyle::drawPrimitive(	PrimitiveElement pe,
					QPainter *p,
					const QRect &r,
					const QColorGroup &cg,
					SFlags flags,
					const QStyleOption &data) const
{
    switch (pe) {
    case PE_HeaderSection:
    {
	flags = ((flags | Style_Sunken) ^ Style_Sunken) | Style_Raised;
        if(isKicker)
            drawLightBevel(p, r, flags, flags & Style_Down ? &gray[2] : &gray[6], gray);
        else
            drawLightBevel(p, r, flags, flags & Style_Down ? &button[3] : &button[1], button);
	break;
    }
    case PE_HeaderArrow:
        drawArrow(p, r, flags & Style_Enabled ? cg.buttonText() : cg.mid(), (flags & Style_Up) ? PE_ArrowUp : PE_ArrowDown);
	break;
    case PE_ButtonCommand:
    case PE_ButtonBevel:
    case PE_ButtonTool:
    case PE_ButtonDropDown:
    {
        QColor colors[7];
        const QColor *use=button;

        if(cg.button()!=button[6])
        {
            shadeColors(cg.button(), colors);
            use=colors;
        }

        const QColor *fill =
            (flags & (QStyle::Style_Down | QStyle::Style_On | QStyle::Style_Sunken)) ? &use[3] :
            (flags & QStyle::Style_MouseOver) ? &use[1] : &use[6];

	drawLightBevel(p, r, flags, fill, use);
	break;
    }
    case PE_ButtonDefault:
    {
	p->setPen(cg.text());
	p->setBrush(NoBrush);
	p->drawRect(r);
	break;
    }
    case PE_Indicator:
    {
        QColor colors[7];
        const QColor *use=button;

        if(cg.button()!=button[6])
        {
            shadeColors(cg.button(), colors);
            use=colors;
        }

        p->fillRect(r.x(), r.y(), CHECK_SIZE, CHECK_SIZE, flags&Style_Enabled ? cg.base() : cg.background());
        p->setPen(use[4]);
        p->drawLine(r.x()+1, r.y()+CHECK_SIZE-1, r.x()+1, r.y()+1);
        p->drawLine(r.x()+1, r.y()+1, r.x()+CHECK_SIZE-2, r.y()+1);
        p->setPen(use[5]);
        p->drawRect(r.x(), r.y(), CHECK_SIZE, CHECK_SIZE);
        if(flags & QStyle::Style_On)
        {
            QPointArray check;

            check.setPoints(6,
                            r.x()+3, r.y()+5,
                            r.x()+5, r.y()+7,
                            r.x()+9, r.y()+3,
                            r.x()+9, r.y()+5,
                            r.x()+5, r.y()+9,
                            r.x()+3, r.y()+7);
            p->setBrush(flags&Style_Enabled ? cg.text() : cg.mid());
            p->setPen(flags&Style_Enabled ? cg.text() : cg.mid());
            p->drawPolygon(check);
        }
	break;
    }
    case PE_ExclusiveIndicator:
    {
        QPointArray outer,
                    shade;
        QColor colors[7];
        const QColor *use=button;

        if(cg.button()!=button[6])
        {
            shadeColors(cg.button(), colors);
            use=colors;
        }

        outer.setPoints(24,
                        r.x(), r.y()+7, r.x(), r.y()+4, r.x()+1, r.y()+3, r.x()+1, r.y()+2, r.x()+2, r.y()+1, r.x()+3, r.y()+1,
                        r.x()+4, r.y(), r.x()+7, r.y(), r.x()+8, r.y()+1, r.x()+9, r.y()+1, r.x()+10, r.y()+2, r.x()+10, r.y()+3,
                        r.x()+11, r.y()+4, r.x()+11, r.y()+7, r.x()+10, r.y()+8, r.x()+10, r.y()+9, r.x()+9, r.y()+10, r.x()+8, r.y()+10,
                        r.x()+7, r.y()+11, r.x()+4, r.y()+11, r.x()+3, r.y()+10, r.x()+2, r.y()+10, r.x()+1, r.y()+9, r.x()+1, r.y()+8);

        shade.setPoints(11,
                         r.x()+2, r.y()+9, r.x()+2, r.y()+8, r.x()+1, r.y()+7, r.x()+1, r.y()+4, r.x()+2, r.y()+3,
                         r.x()+2, r.y()+2, r.x()+3, r.y()+2, r.x()+4, r.y()+1, r.x()+7, r.y()+1, r.x()+8, r.y()+2, r.x()+9, r.y()+2);

        p->fillRect(r, flags&Style_MouseOver ? cg.background().light(HIGHLIGHT_FACTOR) : cg.background());
        p->setBrush(flags&Style_Enabled ? cg.base() : cg.background());
        p->setPen(flags&Style_Enabled ? cg.base() : cg.background());
        p->drawEllipse(r.x(), r.y(), RADIO_SIZE, RADIO_SIZE);
        p->setPen(use[5]);
        p->drawPolyline(outer);
        p->setPen(use[4]);
        p->drawPolyline(shade);
        if(flags & QStyle::Style_On)
        {
            p->setBrush(flags&Style_Enabled ? cg.text() : cg.mid());
            p->setPen(NoPen);
            p->drawRect(r.x() + 5, r.y() + 4, 2, 4);
            p->drawRect(r.x() + 4, r.y() + 5, 4, 2);
        }
        break;
    }
    case PE_Separator:
        qDrawShadeLine(p, r.left(), r.top(), r.right(), r.bottom(),
            QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]), flags & Style_Sunken, 1, 0);
        break;
    case PE_DockWindowSeparator:
    {
        QPoint p1, p2;
        if ( flags & Style_Horizontal ) {
            p1 = QPoint( r.width()/2, 0 );
            p2 = QPoint( p1.x(), r.height() );
        } else {
            p1 = QPoint( 0, r.height()/2 );
            p2 = QPoint( r.width(), p1.y() );
        }
        qDrawShadeLine(p, p1, p2, QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]), 1, 1, 0 );
	break;
    }
    case PE_Splitter:
        p->setPen(gray[5]);
        p->drawLine(r.x()+1,           r.y(),              r.x()+r.width()-2, r.y());
        p->drawLine(r.x()+r.width()-1, r.y()+1,            r.x()+r.width()-1, r.y()+r.height()-2);
        p->drawLine(r.x()+1,           r.y()+r.height()-1, r.x()+r.width()-2, r.y()+r.height()-1);
        p->drawLine(r.x(),             r.y()+1,            r.x(),             r.y()+r.height()-2);
        p->setPen(gray[0]);
        p->drawLine(r.x()+1,           r.y()+1,            r.x()+r.width()-3, r.y()+1);
        p->drawLine(r.x()+1,           r.y()+2,            r.x()+1,           r.y()+r.height()-3);
	break;
    case PE_DockWindowResizeHandle:
	p->fillRect(r, cg.background());
	if (flags & Style_Horizontal) {
	    p->setPen(cg.highlight().light());
	    p->drawLine(r.left() + 1, r.top() + 1, r.right() - 1, r.top() + 1);
	    p->setPen(cg.highlight());
	    p->drawLine(r.left() + 1, r.top() + 2, r.right() - 1, r.top() + 2);
	    p->setPen(cg.highlight().dark());
	    p->drawLine(r.left() + 1, r.top() + 3, r.right() - 1, r.top() + 3);
	} else {
	    p->setPen(cg.highlight().light());
	    p->drawLine(r.left() + 1, r.top() + 1, r.left() + 1, r.bottom() - 1);
	    p->setPen(cg.highlight());
	    p->drawLine(r.left() + 2, r.top() + 1, r.left() + 2, r.bottom() - 1);
	    p->setPen(cg.highlight().dark());
	    p->drawLine(r.left() + 3, r.top() + 1, r.left() + 3, r.bottom() - 1);
	}
	break;
    case PE_GroupBoxFrame:
        if (data.isDefault())
            break;
        if (flags & (Style_Sunken|Style_Raised))
            qDrawShadeRect(p, r.x(), r.y(), r.width(), r.height(), QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]),
                           flags & Style_Sunken, data.lineWidth(), data.midLineWidth());
        else
            qDrawPlainRect(p, r.x(), r.y(), r.width(), r.height(), gray[5], data.lineWidth());
        break;
    case PE_Panel:
    case PE_PanelLineEdit:
    case PE_WindowFrame:
        qDrawShadePanel(p, r, QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]),
                        flags & Style_Sunken, data.isDefault() ? pixelMetric(PM_DefaultFrameWidth) : data.lineWidth());
        break;
    case PE_PanelPopup:
    case PE_PanelTabWidget:
        if(borderFrame && ((data.isDefault()&& pixelMetric(PM_DefaultFrameWidth)>1) || (!data.isDefault() &&data.lineWidth()>1)))
        {
            p->setPen(gray[5]);
            p->drawRect(r);
            qDrawShadePanel(p, r.x()+1, r.y()+1, r.width()-2, r.height()-2,
                            QColorGroup(gray[4], gray[6], gray[0], gray[4], gray[2], cg.text(), gray[6]),
                            flags & Style_Sunken, data.isDefault() ? pixelMetric(PM_DefaultFrameWidth)-1 : data.lineWidth()-1);
        }
        else
            qDrawShadePanel(p, r, QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]),
                            flags & Style_Sunken, data.isDefault() ? pixelMetric(PM_DefaultFrameWidth) : data.lineWidth());
        break;
    case PE_PanelDockWindow:
    case PE_PanelMenuBar:
        qDrawShadePanel(p, r, QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]), flags & Style_Sunken, 1);
	break;
    case PE_ScrollBarAddLine:
    case PE_ScrollBarSubLine:
    {
        bool down=(flags & (QStyle::Style_Down | QStyle::Style_On | QStyle::Style_Sunken));
        const QColor *fill =
            down ? &button[3] :
            (flags & QStyle::Style_MouseOver) ? &button[1] : &button[6];

        drawLightBevel(p, r, down ? flags : ((flags & Style_Enabled) ? Style_Raised : Style_Default), fill, button);

	if ((pe == PE_ScrollBarAddLine) && (flags & Style_Horizontal))
	    pe = PE_ArrowRight;
	else if(pe == PE_ScrollBarAddLine)
	    pe = PE_ArrowDown;
	else if(flags & Style_Horizontal)
	    pe = PE_ArrowLeft;
	else
	    pe = PE_ArrowUp;

	drawPrimitive(pe, p, r, cg, flags);
	break;
    }
    case PE_ScrollBarSubPage:
    case PE_ScrollBarAddPage:
        p->fillRect(r.x(), r.y(), r.width(), r.height(), gray[2]);
        if(borderButton)
        {
            p->setPen(gray[5]);

            if (flags & Style_Horizontal)
            {
                p->drawLine(r.left(), r.top(), r.right(), r.top());
                p->drawLine(r.left(), r.bottom(), r.right(), r.bottom());
            }
            else
            {
                p->drawLine(r.left(), r.top(), r.left(), r.bottom());
                p->drawLine(r.right(), r.top(), r.right(), r.bottom());
            }
        }
	break;
    case PE_ScrollBarSlider:
    {
        const QColor *fill = (flags & (QStyle::Style_Down | QStyle::Style_MouseOver)) || hoverWidget
                                 ? &button[1] : &button[6];

        drawLightBevel(p, r, ((flags | Style_Down) ^ Style_Down) |
                     ((flags & Style_Enabled) ? Style_Raised : Style_Default), fill, button);
        if((flags & Style_Horizontal && r.width() >= 20) || r.height() >= 20)
            drawLines(p, r, !(flags & Style_Horizontal), 4, 3, button);
	break;
    }
    case PE_FocusRect:
	p->drawWinFocusRect(r);
	break;
    case PE_ArrowUp:
    case PE_ArrowDown:
    case PE_ArrowRight:
    case PE_ArrowLeft:
        drawArrow(p, r, flags & Style_Enabled ? cg.buttonText() : cg.mid(), pe);
	break;
    case PE_SpinWidgetUp:
    case PE_SpinWidgetDown:
    {
        QRect sr(r);
        int   min=PE_SpinWidgetDown==pe ? 15 : 14;

        const QColor *fill =
            (flags & (QStyle::Style_Down | QStyle::Style_On | QStyle::Style_Sunken)) ? &button[3] :
            (flags & QStyle::Style_MouseOver) ? &button[1] : &button[6];

        drawLightBevel(p, sr, flags, fill, button);

        if(PE_SpinWidgetDown==pe)
            sr.setY(sr.y()-1);
        drawArrow(p, sr, flags&Style_Enabled ? cg.buttonText() : cg.mid(), (pe==PE_SpinWidgetUp ? PE_ArrowUp : PE_ArrowDown), 
                  true);
	break;
    }
    default:
	KStyle::drawPrimitive(pe, p, r, cg, flags, data);
	break;
    }
}

void QtCurveStyle::drawKStylePrimitive( KStylePrimitive kpe,
                                          QPainter* p,
                                          const QWidget* widget,
                                          const QRect &r,
                                          const QColorGroup &cg,
                                          SFlags flags,
                                          const QStyleOption &opt ) const
{
    switch ( kpe )
    {
        case KPE_ToolBarHandle:
        case KPE_GeneralHandle:
            p->fillRect(r, gray[6]);
            drawLines(p, r, !(flags & Style_Horizontal), 2, isKicker ? 1 : KPE_ToolBarHandle==kpe ? 4 : 2, gray,
                                                            isKicker ||  KPE_ToolBarHandle!=kpe ? 0 : -2);
            break;
        case KPE_SliderGroove:
            drawSliderGroove(p, r, flags, widget);
            break;
        case KPE_SliderHandle:
            drawSliderHandle(p, r, flags);
            break;
        default:
            KStyle::drawKStylePrimitive( kpe, p, widget, r, cg, flags, opt);
    }
}

void QtCurveStyle::drawControl(ControlElement control,
				  QPainter *p,
				  const QWidget *widget,
				  const QRect &r,
				  const QColorGroup &cg,
				  SFlags flags,
				  const QStyleOption &data) const
{
    if (widget == hoverWidget)
        flags |= Style_MouseOver;

    switch (control) {
   case CE_TabBarTab:
        if(borderFrame)
        {
            QRect tr(r),
                  fr(r);

            tr.addCoords(0, 0,  0, -1);
            fr.addCoords(2, 2, -2, -2);

            if (flags & Style_Selected)
            {
                p->setPen(gray[5]);
                if (tr.left() == 0)
                    p->drawLine(tr.left(), tr.bottom() + 1,
                                tr.left(), tr.top() + 1);
                else
                    p->drawLine(tr.left(), tr.bottom(),
                                tr.left(), tr.top() + 1);
                p->drawLine(tr.left(), tr.top(),
                            tr.right(), tr.top());
                p->drawLine(tr.right(), tr.top() + 1,
                            tr.right(), tr.bottom());

                p->setPen(gray[0]);
                if (tr.left() == 0)
                    p->drawLine(tr.left() + 1, tr.bottom() + 2,
                                tr.left() + 1, tr.top() + 2);
                else {
                    p->drawLine(tr.left() + 1, tr.bottom()+1,
                                tr.left() + 1, tr.top() + 2);
                    p->drawPoint(tr.left(), tr.bottom() + 1);
                }
                p->drawLine(tr.left() + 1, tr.top() + 1,
                            tr.right() - 1, tr.top() + 1);
                p->drawPoint(tr.right()-1, tr.bottom()+1);

                p->setPen(gray[4]);
                p->drawLine(tr.right() - 1, tr.top() + 1,
                            tr.right() - 1, tr.bottom());
            }
            else
            {
                tr.addCoords(0, 1, 0, 0);
                fr.addCoords(0, 1, 0, 0);

                p->setPen(gray[5]);
                p->drawRect(tr);

                if (r.left() == 0)
                    p->drawPoint(tr.left(), tr.bottom() + 1);

                p->setPen(gray[0]);
                p->drawLine(tr.left() + 1, tr.top() + 1,
                            tr.right() - 1, tr.top() + 1);
                if(0!=r.left())
                    p->setPen(gray[2]);

                p->drawLine(tr.left() + 1, tr.bottom() - 1,
                            tr.left() + 1, tr.top() + 2);

                if (tr.left() == 0)
                    p->drawLine(tr.left() + 1, tr.bottom() + 1,
                                tr.right(), tr.bottom() + 1);
                else
                {
                    p->setPen(gray[0]);
                    p->drawLine(tr.left(), tr.bottom() + 1,
                                tr.right(), tr.bottom() + 1);
                }

                p->setPen(gray[4]);
                p->drawLine(tr.right() - 1, tr.top() + 1,
                            tr.right() - 1, tr.bottom() - 1);
            }

            p->fillRect(fr, ((flags & Style_Selected) ? 
                             cg.background() : gray[2]));
        }
        else
        {
          const QTabBar *tb = (const QTabBar *) widget;
          QRect br(r);

          switch(tb->shape())
          {
            case QTabBar::RoundedAbove:
                if ( flags & Style_Selected )
                {
                    p->setPen( cg.background() );
                    p->drawLine( br.bottomLeft(), br.bottomRight() );
                    p->setPen( gray[0] );
                    p->drawPoint( br.bottomLeft() );
                    p->setPen( gray[5] );
                    p->drawPoint( br.bottomRight() );
                    br.addCoords( 0, 0, 0, -1 );
                }
                else
                {
                    p->setPen( gray[0] );
                    p->drawLine( br.left(),  br.bottom(),
                                 br.right(), br.bottom() );
                    br.addCoords( 0, 1, -1, -1 );
                }

                p->setPen(gray[0==r.left() || flags & Style_Selected ? 0 : 5]);
                p->drawLine( br.bottomLeft(), br.topLeft() );
                p->setPen( gray[0] );
                p->drawLine( br.left()+1, br.top(), br.right()-1, br.top() );
                p->setPen( gray[5] );
                p->drawLine( br.right(), br.top(), br.right(), br.bottom() );
                br.addCoords( 1, 1, -1, 0 );
                p->fillRect( br, flags & Style_Selected ? cg.background() : gray[2] );
                break;
            case QTabBar::RoundedBelow:
                if ( flags & Style_Selected )
                {
                    p->setPen( cg.background() );
                    p->drawLine( br.topLeft(), br.topRight() );
                    p->setPen( gray[0] );
                    p->drawPoint( br.topLeft() );
                    p->setPen( gray[5] );
                    p->drawPoint( br.topRight() );
                    br.addCoords( 0, 1, 0, 0 );
                }
                else
                {
                    p->setPen( gray[5] );
                    p->drawLine( br.left(),  br.top(),
                                 br.right(), br.top() );
                    br.addCoords( 0, 1, -1, -1 );
                }

                if(0==r.left() || flags & Style_Selected)
                {
                  p->setPen(gray[0]);
                  p->drawLine( br.bottomLeft(), br.topLeft() );
                }
                p->setPen( gray[5] );
                p->drawLine( br.bottomLeft(), br.bottomRight() );
                p->drawLine( br.right(), br.top(), br.right(), br.bottom() );
                br.addCoords( 1, 0, -1, -1 );
                p->fillRect( br, flags & Style_Selected ? cg.background() : gray[2] );
                break;
            default:
                KStyle::drawControl( control, p, widget, r, cg, flags, data );
          }
        }
        break;
    case CE_PopupMenuItem:
    {
	if (! widget || data.isDefault())
	    break;

	const QPopupMenu *popupmenu = (const QPopupMenu *) widget;
	QMenuItem *mi = data.menuItem();
	int tab = data.tabWidth();
	int maxpmw = data.maxIconWidth();
        int x, y, w, h;
        r.rect( &x, &y, &w, &h );

        if ((flags & Style_Active) && (flags & Style_Enabled))
            drawGradient(p, r);
        else if ( widget->erasePixmap() && !widget->erasePixmap()->isNull() )
            p->drawPixmap( x, y, *widget->erasePixmap(), x, y, w, h );
        else
            p->fillRect(r, gray[6]);

	if ( !mi )
	    break;

        if ( mi->isSeparator() ) {
            p->setPen(gray[5]);
            p->drawLine(r.left(), r.top(), r.right(), r.top());
            p->setPen(gray[0]);
            p->drawLine(r.left(), r.top()+1, r.right(), r.top()+1);
            break;
        }

	maxpmw = QMAX(maxpmw, 16);

	QRect cr, ir, tr, sr;
	// check column
	cr.setRect(r.left(), r.top(), maxpmw, r.height());
	// submenu indicator column
	sr.setCoords(r.right() - maxpmw, r.top(), r.right(), r.bottom());
	// tab/accelerator column
	tr.setCoords(sr.left() - tab - 4, r.top(), sr.left(), r.bottom());
	// item column
	ir.setCoords(cr.right() + 4, r.top(), tr.right() - 4, r.bottom());

	bool reverse = QApplication::reverseLayout();
	if ( reverse ) {
	    cr = visualRect( cr, r );
	    sr = visualRect( sr, r );
	    tr = visualRect( tr, r );
	    ir = visualRect( ir, r );
	}

	if (mi->iconSet()) {
            // Select the correct icon from the iconset
            QIconSet::Mode mode =  flags & Style_Active
                                      ? (mi->isEnabled() ? QIconSet::Active : QIconSet::Disabled)
                                      : (mi->isEnabled() ? QIconSet::Normal : QIconSet::Disabled);
            cr = visualRect( QRect(x, y, maxpmw, h), r );

            // Do we have an icon and are checked at the same time?
            // Then draw a "pressed" background behind the icon
            if ( popupmenu->isCheckable() && mi->isChecked() )
            {
                QBrush brush(gray[3]);
                qDrawShadePanel(p, cr.x()+1, cr.y()+2, cr.width()-2, cr.height()-4,
                                QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]),
                                true, 1, &brush);
            }
            // Draw the icon
            QPixmap pixmap = mi->iconSet()->pixmap( QIconSet::Small, mode );
            QRect pmr( 0, 0, pixmap.width(), pixmap.height() );
            pmr.moveCenter( cr.center() );
            p->drawPixmap( pmr.topLeft(), pixmap );
	} else if (popupmenu->isCheckable() && mi->isChecked())
	    drawPrimitive(PE_CheckMark, p, cr, cg,
			  (flags & (Style_Enabled|Style_Active)) | Style_On);

        QColor textcolor;
        QColor embosscolor;
        if (flags & Style_Active) {
            if (! (flags & Style_Enabled)) {
                textcolor = cg.text();
                embosscolor = cg.light();
            } else {
                textcolor = cg.highlightedText();
                embosscolor = cg.midlight().light();
            }
        } else if (! (flags & Style_Enabled)) {
            textcolor = cg.text();
            embosscolor = cg.light();
        } else
            textcolor = embosscolor = cg.buttonText();
        p->setPen(textcolor);

	if (mi->custom()) {
	    p->save();
	    if (! (flags & Style_Enabled)) {
		p->setPen(cg.light());
		mi->custom()->paint(p, cg, (flags & Style_Enabled) ? (flags & Style_Active) : 0,
				    flags & Style_Enabled,
				    ir.x() + 1, ir.y() + 1,
				    ir.width() - 1, ir.height() - 1);
		p->setPen(textcolor);
	    }
	    mi->custom()->paint(p, cg, (flags & Style_Enabled) ? (flags & Style_Active) : 0,
				flags & Style_Enabled,
				ir.x(), ir.y(),
				ir.width(), ir.height());
	    p->restore();
	}

	QString text = mi->text();
	if (! text.isNull()) {
	    int t = text.find('\t');

	    // draw accelerator/tab-text
	    if (t >= 0) {
		int alignFlag = AlignVCenter | ShowPrefix | DontClip | SingleLine;
		alignFlag |= ( reverse ? AlignLeft : AlignRight );
		if (! (flags & Style_Enabled)) {
		    p->setPen(embosscolor);
		    tr.moveBy(1, 1);
		    p->drawText(tr, alignFlag, text.mid(t + 1));
		    tr.moveBy(-1, -1);
		    p->setPen(textcolor);
		}

		p->drawText(tr, alignFlag, text.mid(t + 1));
	    }

	    int alignFlag = AlignVCenter | ShowPrefix | DontClip | SingleLine;
	    alignFlag |= ( reverse ? AlignRight : AlignLeft );

	    if (! (flags & Style_Enabled)) {
		p->setPen(embosscolor);
		ir.moveBy(1, 1);
		p->drawText(ir, alignFlag, text, t);
		ir.moveBy(-1, -1);
		p->setPen(textcolor);
	    }

	    p->drawText(ir, alignFlag, text, t);
	} else if (mi->pixmap()) {
	    QPixmap pixmap = *mi->pixmap();
	    if (pixmap.depth() == 1)
		p->setBackgroundMode(OpaqueMode);
	    p->drawPixmap(ir.x(), (ir.height() - pixmap.height()) / 2, pixmap);
	    if (pixmap.depth() == 1)
		p->setBackgroundMode(TransparentMode);
	}

	if (mi->popup())
            drawArrow(p, sr, textcolor, reverse ? PE_ArrowLeft : PE_ArrowRight);
	break;
    }
    case CE_MenuBarItem:
    {
	if ((flags & Style_Enabled) && (flags & Style_Active) && (flags & Style_Down))
	    drawGradient(p, r);

	if (data.isDefault())
	    break;

	QMenuItem *mi = data.menuItem();
	if (flags & Style_Active && (flags & Style_Down))
	    drawItem(p, r, AlignCenter | ShowPrefix | DontClip | SingleLine, cg, flags & Style_Enabled, mi->pixmap(), mi->text(), -1, &cg.highlightedText());
	else
	    drawItem(p, r, AlignCenter | ShowPrefix | DontClip | SingleLine, cg, flags & Style_Enabled, mi->pixmap(), mi->text(), -1, &cg.buttonText());
	break;
    }
    case CE_MenuBarEmptyArea:
    case CE_DockWindowEmptyArea:
        p->fillRect(r, cg.background());
        break;
    case CE_ProgressBarGroove:
	p->setBrush(gray[6]);
	p->drawRect(r);
        qDrawShadePanel(p, r, QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]), true, 1);
	break;
    case CE_ProgressBarContents:
    {
        // ### Take into account totalSteps() for busy indicator
        const QProgressBar* pb = (const QProgressBar*)widget;
        QRect cr = subRect(SR_ProgressBarContents, widget);
        double progress = pb->progress();
        bool reverse = QApplication::reverseLayout();
        int steps = pb->totalSteps();

        if (cr.isValid() && (progress > 0 || steps == 0)) 
        {
            double pg = (steps == 0) ? 0.1 : progress / steps;
            int width = QMIN(cr.width(), (int)(pg * cr.width()));
            if (steps == 0) //Busy indicator
            {
                if (width < 1) width = 1; //A busy indicator with width 0 is kind of useless

                int remWidth = cr.width() - width; //Never disappear completely
                if (remWidth <= 0) remWidth = 1; //Do something non-crashy when too small...

                int pstep =  int(progress) % ( 2 *  remWidth );

                if (pstep > remWidth)
                {
                    //Bounce about.. We're remWidth + some delta, we want to be remWidth - delta...
                    // - ( (remWidth + some delta) - 2* remWidth )  = - (some deleta - remWidth) = remWidth - some delta..
                    pstep = - (pstep - 2 * remWidth );
                }

                if (reverse)
                    drawGradient(p, QRect(cr.x() + cr.width() - width - pstep, cr.y(), width, cr.height()));
                else
                    drawGradient(p, QRect(cr.x() + pstep, cr.y(), width, cr.height()));
            }
            else
                if (reverse)
                    drawGradient(p, QRect(cr.x()+(cr.width()-width), cr.y(), width, cr.height()));
                else
                    drawGradient(p, QRect(cr.x(), cr.y(), width, cr.height()));
        }
        break;
    }
    case CE_CheckBox:
	drawPrimitive(PE_Indicator, p, r, cg, flags, data);
	break;
    case CE_CheckBoxLabel:
      {
          const QCheckBox *checkbox = (const QCheckBox *) widget;
          
          if (flags & Style_MouseOver) {
              QRegion r(checkbox->rect());
              r -= visualRect(subRect(SR_CheckBoxIndicator, widget), widget);
              p->setClipRegion(r);
              p->fillRect(checkbox->rect(), cg.background().light(HIGHLIGHT_FACTOR)); // cg.brush(QColorGroup::Midlight));
              p->setClipping(false);
          }
          int alignment = QApplication::reverseLayout() ? AlignRight : AlignLeft;
          drawItem(p, r, alignment | AlignVCenter | ShowPrefix, cg,
                   flags & Style_Enabled, checkbox->pixmap(), checkbox->text());
          
          if (checkbox->hasFocus()) {
              QRect fr = visualRect(subRect(SR_CheckBoxFocusRect, widget), widget);
              drawPrimitive(PE_FocusRect, p, fr, cg, flags);
          }
      }
      break;
    case CE_RadioButtonLabel:
      {
	const QRadioButton *radiobutton = (const QRadioButton *) widget;
	
	if (flags & Style_MouseOver) {
	  QRegion r(radiobutton->rect());
	  r -= visualRect(subRect(SR_RadioButtonIndicator, widget), widget);
	  p->setClipRegion(r);
	  p->fillRect(radiobutton->rect(), cg.background().light(HIGHLIGHT_FACTOR)); // cg.brush(QColorGroup::Midlight));
	  p->setClipping(false);
	}
	
	int alignment = QApplication::reverseLayout() ? AlignRight : AlignLeft;
	drawItem(p, r, alignment | AlignVCenter | ShowPrefix, cg,
		 flags & Style_Enabled, radiobutton->pixmap(), radiobutton->text());
	
	if (radiobutton->hasFocus()) {
	  QRect fr = visualRect(subRect(SR_RadioButtonFocusRect, widget), widget);
	  drawPrimitive(PE_FocusRect, p, fr, cg, flags);
	}
      }
      break;
    default:
	KStyle::drawControl(control, p, widget, r, cg, flags, data);
	break;
   }
}

void QtCurveStyle::drawControlMask(ControlElement control,
				      QPainter *p,
				      const QWidget *widget,
				      const QRect &r,
				      const QStyleOption &data) const
{
    switch (control) {
    case CE_PushButton:
	p->fillRect(r, color1);
	break;
    default:
	KStyle::drawControlMask(control, p, widget, r, data);
	break;
    }
}

QRect QtCurveStyle::subRect(SubRect subrect, const QWidget *widget) const
{
    QRect rect, wrect(widget->rect());

    switch (subrect) {
    case SR_PushButtonFocusRect:
    {
	const QPushButton *button = (const QPushButton *) widget;
	int dbw1 = 0, dbw2 = 0;
	if (button->isDefault() || button->autoDefault()) {
	    dbw1 = pixelMetric(PM_ButtonDefaultIndicator, widget);
	    dbw2 = dbw1 * 2;
	}

	rect.setRect(wrect.x()	  + 3 + dbw1,
		     wrect.y()	  + 3 + dbw1,
		     wrect.width()  - 6 - dbw2,
		     wrect.height() - 6 - dbw2);
	break;
    }
    case SR_CheckBoxIndicator:
        {
            int h = pixelMetric( PM_IndicatorHeight );
            rect.setRect(( widget->rect().height() - h ) / 2,
                         ( widget->rect().height() - h ) / 2,
                         pixelMetric( PM_IndicatorWidth ), h );
            break;
        }
    case SR_RadioButtonIndicator:
        {
            int h = pixelMetric( PM_ExclusiveIndicatorHeight );
            rect.setRect( ( widget->rect().height() - h ) / 2,
                          ( widget->rect().height() - h ) / 2,
                          pixelMetric( PM_ExclusiveIndicatorWidth ), h );
            break;
        }
    case SR_ProgressBarContents:
        rect=QRect(wrect.x()+1, wrect.y()+1, wrect.width()-2, wrect.height()-2);
        break; 
    default:
	rect = KStyle::subRect(subrect, widget);
    }

    return rect;
}

void QtCurveStyle::drawComplexControl(ComplexControl control,
					 QPainter* p,
					 const QWidget* widget,
					 const QRect& r,
					 const QColorGroup& cg,
					 SFlags flags,
					 SCFlags controls,
					 SCFlags active,
					 const QStyleOption &data) const
{
    if (widget == hoverWidget)
        flags |= Style_MouseOver;

    switch (control) {
    case CC_ToolButton:
    {
            const QToolButton *toolbutton = (const QToolButton *) widget;

            QRect button, menuarea;
            button   = querySubControlMetrics(control, widget, SC_ToolButton, data);
            menuarea = querySubControlMetrics(control, widget, SC_ToolButtonMenu, data);

            SFlags bflags = flags,
                   mflags = flags;

            if (active & SC_ToolButton)
                bflags |= Style_Down;
            if (active & SC_ToolButtonMenu)
                mflags |= Style_Down;

            if (controls & SC_ToolButton)
            {
                // If we're pressed, on, or raised...
                if (bflags & (Style_Down | Style_On | Style_Raised))
                    drawPrimitive(PE_ButtonTool, p, button, cg, bflags, data);

                // Check whether to draw a background pixmap
                else if ( toolbutton->parentWidget() &&
                          toolbutton->parentWidget()->backgroundPixmap() &&
                          !toolbutton->parentWidget()->backgroundPixmap()->isNull() )
                {
                    QPixmap pixmap = *(toolbutton->parentWidget()->backgroundPixmap());
                    p->drawTiledPixmap( r, pixmap, toolbutton->pos() );
                }
                else if (widget->parent())
                {
                    if (widget->parent()->inherits("QToolBar"))
                    {
                        QToolBar* parent = (QToolBar*)widget->parent();
                        QRect pr = parent->rect();
                        p->fillRect(pr, cg.background());
                    }
                    else if (widget->parent()->inherits("QToolBarExtensionWidget"))
                    {
                        QWidget* parent = (QWidget*)widget->parent();
                        QToolBar* toolbar = (QToolBar*)parent->parent();
                        QRect tr = toolbar->rect();

                        p->fillRect(tr, cg.background());
                    }
                }
            }

            if (controls & SC_ToolButtonMenu)
            {
                if (mflags & (Style_Down | Style_On | Style_Raised))
                    drawPrimitive(PE_ButtonDropDown, p, menuarea, cg, mflags, data);
                drawPrimitive(PE_ArrowDown, p, menuarea, cg, mflags, data);
            }

            if (toolbutton->hasFocus() && !toolbutton->focusProxy()) {
                QRect fr = toolbutton->rect();
                fr.addCoords(3, 3, -3, -3);
                drawPrimitive(PE_FocusRect, p, fr, cg);
            }

            break;
 
	}
    case CC_ListView:
    {
	if ( controls & SC_ListView )
	    KStyle::drawComplexControl( control, p, widget, r, cg, flags, controls, active, data );
	if ( controls & ( SC_ListViewBranch | SC_ListViewExpand ) ) {
	    if (data.isDefault())
		break;

	    QListViewItem *item = data.listViewItem(),
		*child = item->firstChild();

	    int y = r.y();
	    int c;
	    int dotoffset = 0;
	    QPointArray dotlines;
	    if ( active == SC_All && controls == SC_ListViewExpand ) {
		c = 2;
		dotlines.resize(2);
		dotlines[0] = QPoint( r.right(), r.top() );
		dotlines[1] = QPoint( r.right(), r.bottom() );
	    } else {
		int linetop = 0, linebot = 0;
		// each branch needs at most two lines, ie. four end points
		dotoffset = (item->itemPos() + item->height() - y) %2;
		dotlines.resize( item->childCount() * 4 );
		c = 0;

		// skip the stuff above the exposed rectangle
		while ( child && y + child->height() <= 0 ) {
		    y += child->totalHeight();
		    child = child->nextSibling();
		}

		int bx = r.width() / 2;

		// paint stuff in the magical area
		QListView* v = item->listView();
		int lh;
		if ( !item->multiLinesEnabled() )
		    lh = child ? child->height() : 0;
		else
		    lh = p->fontMetrics().height() + 2 * v->itemMargin();
		lh = QMAX( lh, QApplication::globalStrut().height() );
		if ( lh % 2 > 0 )
		    lh++;
		while ( child && y < r.height() ) {
		    linebot = y + lh/2;
		    if ( (child->isExpandable() || child->childCount()) &&
			 (child->height() > 0) ) {
			// needs a box
			p->setPen( cg.mid() );
			p->drawRect( bx-4, linebot-4, 9, 9 );
			// plus or minus
			p->setPen( cg.text() );
			p->drawLine( bx - 2, linebot, bx + 2, linebot );
			if ( !child->isOpen() )
			    p->drawLine( bx, linebot - 2, bx, linebot + 2 );
			// dotlinery
			p->setPen( cg.mid() );
			dotlines[c++] = QPoint( bx, linetop );
			dotlines[c++] = QPoint( bx, linebot - 4 );
			dotlines[c++] = QPoint( bx + 5, linebot );
			dotlines[c++] = QPoint( r.width(), linebot );
			linetop = linebot + 5;
		    } else {
			// just dotlinery
			dotlines[c++] = QPoint( bx+1, linebot );
			dotlines[c++] = QPoint( r.width(), linebot );
		    }

		    y += child->totalHeight();
		    child = child->nextSibling();
		}

		if ( child ) // there's a child, so move linebot to edge of rectangle
		    linebot = r.height();

		if ( linetop < linebot ) {
		    dotlines[c++] = QPoint( bx, linetop );
		    dotlines[c++] = QPoint( bx, linebot );
		}
	    }
	    p->setPen( cg.text() );

	    static QBitmap *verticalLine = 0, *horizontalLine = 0;
	    static QCleanupHandler<QBitmap> qlv_cleanup_bitmap;
	    if ( !verticalLine ) {
		// make 128*1 and 1*128 bitmaps that can be used for
		// drawing the right sort of lines.
		verticalLine = new QBitmap( 1, 129, true );
		horizontalLine = new QBitmap( 128, 1, true );
		QPointArray a( 64 );
		QPainter p;
		p.begin( verticalLine );
		int i;
		for( i=0; i<64; i++ )
		    a.setPoint( i, 0, i*2+1 );
		p.setPen( color1 );
		p.drawPoints( a );
		p.end();
		QApplication::flushX();
		verticalLine->setMask( *verticalLine );
		p.begin( horizontalLine );
		for( i=0; i<64; i++ )
		    a.setPoint( i, i*2+1, 0 );
		p.setPen( color1 );
		p.drawPoints( a );
		p.end();
		QApplication::flushX();
		horizontalLine->setMask( *horizontalLine );
		qlv_cleanup_bitmap.add( &verticalLine );
		qlv_cleanup_bitmap.add( &horizontalLine );
	    }

	    int line; // index into dotlines
	    for( line = 0; line < c; line += 2 ) {
		// assumptions here: lines are horizontal or vertical.
		// lines always start with the numerically lowest
		// coordinate.

		// point ... relevant coordinate of current point
		// end ..... same coordinate of the end of the current line
		// other ... the other coordinate of the current point/line
		if ( dotlines[line].y() == dotlines[line+1].y() ) {
		    int end = dotlines[line+1].x();
		    int point = dotlines[line].x();
		    int other = dotlines[line].y();
		    while( point < end ) {
			int i = 128;
			if ( i+point > end )
			    i = end-point;
			p->drawPixmap( point, other, *horizontalLine,
				       0, 0, i, 1 );
			point += i;
		    }
		} else {
		    int end = dotlines[line+1].y();
		    int point = dotlines[line].y();
		    int other = dotlines[line].x();
		    int pixmapoffset = ((point & 1) != dotoffset ) ? 1 : 0;
		    while( point < end ) {
			int i = 128;
			if ( i+point > end )
			    i = end-point;
			p->drawPixmap( other, point, *verticalLine,
				       0, pixmapoffset, 1, i );
			point += i;
		    }
		}
	    }
	}
	break;
    }
    case CC_ComboBox:
    {
	const QComboBox *combobox = (const QComboBox *) widget;
	QRect frame(QStyle::visualRect(querySubControlMetrics(CC_ComboBox, widget, SC_ComboBoxFrame, data), widget)),
	      arrow(QStyle::visualRect(querySubControlMetrics(CC_ComboBox, widget, SC_ComboBoxArrow, data), widget)),
	      field(QStyle::visualRect(querySubControlMetrics(CC_ComboBox, widget, SC_ComboBoxEditField, data), widget));

        const QColor *fill =
              (flags & (QStyle::Style_Down | QStyle::Style_On | QStyle::Style_Sunken)) ? &button[3] :
              (flags & QStyle::Style_MouseOver) ? &button[1] : &button[6];

	if ((controls & SC_ComboBoxFrame) && frame.isValid())
            drawLightBevel(p, r, flags | Style_Raised, fill, button);

	if ((controls & SC_ComboBoxArrow) && arrow.isValid()) {
	    if (active == SC_ComboBoxArrow)
                p->fillRect(arrow, *fill);
	    drawPrimitive(PE_ArrowDown, p, arrow, cg, flags & ~Style_MouseOver);
	}

	if ((controls & SC_ComboBoxEditField) && field.isValid()) {
	    if (combobox->editable()) {
              field.addCoords(-1, -1, 1, 1);
                  qDrawShadePanel(p, field,
                        QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]), true, 1);
	    } else {
              field.addCoords(0, -2, 0, 2);
              p->setPen(button[borderButton ? 4 : 5]);
              p->drawLine(field.right(), field.top(), field.right(), field.bottom());
              p->setPen(button[0]);
              p->drawLine(field.right()+1, field.top(), field.right() + 1, field.bottom());
            }

	    if ((flags & Style_HasFocus) && (! combobox->editable()))
            {
                QRect fr=QStyle::visualRect(subRect(SR_ComboBoxFocusRect, widget), widget );

                fr.addCoords(0, 0, -2, 0);
                drawPrimitive(PE_FocusRect, p, fr, cg, flags | Style_FocusAtBorder, QStyleOption(cg.highlight()));
	    }
            if (flags & Style_Enabled)
                p->setPen(cg.buttonText());
            else
                p->setPen(cg.mid());
	}

	break;
    }
    case CC_SpinWidget:
    {
	const QSpinWidget *spinwidget = (const QSpinWidget *) widget;
	QRect frame(querySubControlMetrics(CC_SpinWidget, widget, SC_SpinWidgetFrame, data)),
              up(spinwidget->upRect()),
              down(spinwidget->downRect());

	if ((controls & SC_SpinWidgetFrame) && frame.isValid())
            qDrawShadePanel(p, r, QColorGroup(gray[5], gray[6], gray[0], gray[5], gray[2], cg.text(), gray[6]),
                            true, pixelMetric(PM_SpinBoxFrameWidth));

	if ((controls & SC_SpinWidgetUp) && up.isValid()) {
	    PrimitiveElement pe = PE_SpinWidgetUp;
            SFlags           upflags=flags;

	    if (spinwidget->buttonSymbols() == QSpinWidget::PlusMinus)
		pe = PE_SpinWidgetPlus;
            if(!spinwidget->isUpEnabled())
                upflags^=Style_Enabled;
            drawPrimitive(pe, p, up, cg, upflags | ((active == SC_SpinWidgetUp) ? Style_On | Style_Sunken : Style_Raised));
	}

	if ((controls & SC_SpinWidgetDown) && down.isValid()) {
	    PrimitiveElement pe = PE_SpinWidgetDown;
            SFlags           downflags=flags;

	    if (spinwidget->buttonSymbols() == QSpinWidget::PlusMinus)
		pe = PE_SpinWidgetMinus;
            if(!spinwidget->isDownEnabled())
                downflags^=Style_Enabled;
            drawPrimitive(pe, p, down, cg, downflags | ((active == SC_SpinWidgetDown) ? Style_On | Style_Sunken : Style_Raised));
	}

	break;
    }
    case CC_ScrollBar:
    {
	const QScrollBar *scrollbar = (const QScrollBar *) widget;
	QRect subline(querySubControlMetrics(control, widget, SC_ScrollBarSubLine, data)),
	      addline(querySubControlMetrics(control, widget, SC_ScrollBarAddLine, data)),
	      subpage(querySubControlMetrics(control, widget, SC_ScrollBarSubPage, data)),
	      addpage(querySubControlMetrics(control, widget, SC_ScrollBarAddPage, data)),
	      slider (querySubControlMetrics(control, widget, SC_ScrollBarSlider,  data)),
	      first  (querySubControlMetrics(control, widget, SC_ScrollBarFirst,   data)),
	      last   (querySubControlMetrics(control, widget, SC_ScrollBarLast,    data));

	if ((controls & SC_ScrollBarSubLine) && subline.isValid())
	    drawPrimitive(PE_ScrollBarSubLine, p, subline, cg, Style_Enabled | ((active == SC_ScrollBarSubLine) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
	if ((controls & SC_ScrollBarAddLine) && addline.isValid())
	    drawPrimitive(PE_ScrollBarAddLine, p, addline, cg, Style_Enabled | ((active == SC_ScrollBarAddLine) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
	if ((controls & SC_ScrollBarSubPage) && subpage.isValid())
	    drawPrimitive(PE_ScrollBarSubPage, p, subpage, cg, Style_Enabled | ((active == SC_ScrollBarSubPage) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
	if ((controls & SC_ScrollBarAddPage) && addpage.isValid())
	    drawPrimitive(PE_ScrollBarAddPage, p, addpage, cg, ((scrollbar->minValue() == scrollbar->maxValue()) ? Style_Default : Style_Enabled) | ((active == SC_ScrollBarAddPage) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
	if ((controls & SC_ScrollBarFirst) && first.isValid())
	    drawPrimitive(PE_ScrollBarFirst, p, first, cg, Style_Enabled | ((active == SC_ScrollBarFirst) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
	if ((controls & SC_ScrollBarLast) && last.isValid())
	    drawPrimitive(PE_ScrollBarLast, p, last, cg, Style_Enabled | ((active == SC_ScrollBarLast) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
	if ((controls & SC_ScrollBarSlider) && slider.isValid()) {
	    drawPrimitive(PE_ScrollBarSlider, p, slider, cg, Style_Enabled | ((active == SC_ScrollBarSlider) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));

	    // ### perhaps this should not be able to accept focus if maxedOut?
	    if (scrollbar->hasFocus()) {
		QRect fr(slider.x() + 2, slider.y() + 2, slider.width() - 5, slider.height() - 5);
		drawPrimitive(PE_FocusRect, p, fr, cg, Style_Default);
	    }
	}

	break;
    }
    case CC_Slider:
    {
        QRect groove = querySubControlMetrics(CC_Slider, widget, SC_SliderGroove, data),
              handle = querySubControlMetrics(CC_Slider, widget, SC_SliderHandle, data);

        if ((controls & SC_SliderGroove) && groove.isValid())
            drawSliderGroove(p, groove, flags, widget);
        if ((controls & SC_SliderHandle) && handle.isValid())
            drawSliderHandle(p, handle, flags);
        if (controls & SC_SliderTickmarks)
            QCommonStyle::drawComplexControl(control, p, widget, r, cg, flags,
                                             SC_SliderTickmarks, active, data );
        break;
    }
    default:
	KStyle::drawComplexControl(control, p, widget, r, cg, flags,
					 controls, active, data);
	break;
    }
}

QRect QtCurveStyle::querySubControlMetrics( ComplexControl control,
					       const QWidget *widget,
					       SubControl sc,
					       const QStyleOption &data ) const
{
    switch (control)
    {
        case CC_SpinWidget:
        {
            if ( !widget )
                return QRect();
            int fw = pixelMetric( PM_SpinBoxFrameWidth, 0 );
            QSize bs;
            bs.setHeight( widget->height()/2 );
            if (bs.height() < 8)
                bs.setHeight(8);
            bs.setWidth(QMIN(bs.height()*8/6, widget->width()/4));
            bs = bs.expandedTo( QApplication::globalStrut() );

            int extra=bs.height()*2==widget->height() ? 0 : 1;
            int y=0,
                x=widget->width()-y-bs.width(),
                lx=fw,
                rx=x-fw*2;
            switch ( sc )
            {
                case SC_SpinWidgetUp:
                    return QRect(x, y, bs.width(), bs.height());
                case SC_SpinWidgetDown:
                    return QRect(x, y+bs.height(), bs.width(), bs.height()+extra);
                case SC_SpinWidgetButtonField:
                    return QRect(x, y, bs.width(), widget->height() - 2*fw);
                case SC_SpinWidgetEditField:
                    return QRect(lx, fw, rx, widget->height() - 2*fw);
                case SC_SpinWidgetFrame:
                    return QRect(widget->x(), widget->y(), widget->width() - bs.width(), widget->height());
            }
        }
        default:
	    return KStyle::querySubControlMetrics(control, widget, sc, data);
    }
}

QStyle::SubControl QtCurveStyle::querySubControl( ComplexControl control,
						     const QWidget *widget,
						     const QPoint &pos,
						     const QStyleOption &data ) const
{
    QStyle::SubControl ret = KStyle::querySubControl(control, widget, pos, data);

    return ret;
}

int QtCurveStyle::pixelMetric( PixelMetric metric, const QWidget *widget ) const
{
    switch (metric) {
    case PM_ButtonMargin:
	return 8;
    case PM_ButtonShiftHorizontal:
    case PM_ButtonShiftVertical:
	return 1;
    case PM_ButtonDefaultIndicator:
	return 1;
    case PM_DefaultFrameWidth:
	return !borderFrame || (widget && (widget->inherits("QLineEdit") || widget->inherits("QTextView") ||
                                widget->inherits("QScrollView"))) ? 1 : 2;
    case PM_SpinBoxFrameWidth:
        return 1;
    case PM_IndicatorWidth:
    case PM_IndicatorHeight:
        return CHECK_SIZE;
    case PM_ExclusiveIndicatorWidth:
    case PM_ExclusiveIndicatorHeight:
	return RADIO_SIZE;
    case PM_TabBarTabOverlap:
	return borderFrame ? 1 : 2;
    case PM_ScrollBarExtent:
	return 15;
    case PM_ProgressBarChunkWidth:
	return 2;
    case PM_DockWindowSeparatorExtent:
	return 4;
    case PM_DockWindowHandleExtent:
	return 10;
    case PM_SplitterWidth:
	return 6;
    case PM_ScrollBarSliderMin:
	return 16;
    case PM_SliderControlThickness:
	return 14;
    case PM_SliderThickness:
        return 14;
    case PM_SliderLength:
	return 24;
    case PM_MaximumDragDistance:
	return -1;
    default:
	return KStyle::pixelMetric(metric, widget);
    }
}

int QtCurveStyle::kPixelMetric(KStylePixelMetric kpm, const QWidget *widget) const
{
    switch(kpm)
    {
        case KPM_MenuItemSeparatorHeight:
            return 2;
        default:
            return KStyle::kPixelMetric(kpm, widget);
    }
}


QSize QtCurveStyle::sizeFromContents( ContentsType contents,
					 const QWidget *widget,
					 const QSize &contentsSize,
					 const QStyleOption &data ) const
{
    QSize ret;

    switch (contents) {
    case CT_PushButton:
    {
	const QPushButton *button = (const QPushButton *) widget;
	ret = KStyle::sizeFromContents( contents, widget, contentsSize, data );
	int w = ret.width(), h = ret.height();

	// only expand the button if we are displaying text...
	if ( ! button->pixmap() ) {
	    if ( button->isDefault() || button->autoDefault() ) {
		// default button minimum size
		if ( w < 80 )
		    w = 80;
		if ( h < 25 )
		    h = 25;
	    } else {
		// regular button minimum size
		if ( w < 54 )
		    w = 54;
		if ( h < 21 )
		    h = 21;
	    }
	}

	ret = QSize( w, h );
	break;
    }
    case CT_ComboBox:
    {
	ret = KStyle::sizeFromContents( contents, widget, 
					      contentsSize, data );
	int w = ret.width(), h = ret.height();
	if (h < 27)
	    h = 27;

	ret = QSize(w, h);
	break;
    }
    default:
	ret = KStyle::sizeFromContents(contents, widget, contentsSize, data);
	break;
    }

    return ret;
}

int QtCurveStyle::styleHint( StyleHint stylehint,
				const QWidget *widget,
				const QStyleOption &option,
				QStyleHintReturn* returnData ) const
{
    switch (stylehint) {
    case SH_EtchDisabledText:
    case SH_Slider_SnapToValue:
    case SH_PrintDialog_RightAlignButtons:
    case SH_FontDialog_SelectAssociatedText:
    case SH_MenuBar_AltKeyNavigation:
    case SH_MenuBar_MouseTracking:
    case SH_PopupMenu_MouseTracking:
    case SH_PopupMenu_SpaceActivatesItem:
    case SH_ComboBox_ListMouseTracking:
    case SH_ScrollBar_MiddleClickAbsolutePosition:
	return 1;
    case SH_MainWindow_SpaceBelowMenuBar:
	return 0;
    case SH_PopupMenu_AllowActiveAndDisabled:
	return 0;
    default:
	return KStyle::styleHint(stylehint, widget, option, returnData);
    }
}

void QtCurveStyle::drawGradient(QPainter *p, QRect const &r, bool horiz) const
{
    QRect  grad(r.left()+2, r.top()+2, r.width()-3, r.height()-3);
    QColor col(gradient[GRADIENT_BASE]);
    int    i,
           s=horiz ? grad.top() : grad.left(),
           e=horiz ? grad.bottom() : grad.right();
    double amt=(horiz ? grad.bottom()-grad.top() : grad.right() - grad.left())-1,
           dr, db, dg;

    dr = ((double)(gradient[GRADIENT_BOTTOM].red() -   gradient[GRADIENT_TOP].red())) / amt;
    dg = ((double)(gradient[GRADIENT_BOTTOM].green() - gradient[GRADIENT_TOP].green())) / amt;
    db = ((double)(gradient[GRADIENT_BOTTOM].blue() -  gradient[GRADIENT_TOP].blue())) / amt;

    double rc=0, gc=0, bc=0;

   col=gradient[GRADIENT_TOP];

    for (i = s; i < e; i++)
    //for(i=e-1; i>=s; i--)
    {
      p->setPen(QColor(limit(gradient[GRADIENT_TOP].red()+rc), limit(gradient[GRADIENT_TOP].green()+gc), limit(gradient[GRADIENT_TOP].blue()+bc)));
      if(horiz)
          p->drawLine(grad.left(), i, grad.right(), i);
      else
          p->drawLine(grad.top(), i, grad.bottom(), i);
      rc+=dr; gc+=dg; bc+=db;
    }

    // 3d border effect...
    p->setPen(gradient[GRADIENT_BASE]);
    p->setBrush(NoBrush);
    p->drawRect(r);
    p->setPen(gradient[GRADIENT_LIGHT]);
    p->drawLine(r.left()+1, r.top()+1, r.right()-1, r.top()+1);
    p->drawLine(r.left()+1, r.top()+1, r.left()+1, r.bottom()-1);
    p->setPen(gradient[GRADIENT_DARK]);
    p->drawLine(r.left()+2, r.bottom()-1, r.right()-1, r.bottom()-1);
    p->drawLine(r.right()-1, r.bottom()-1, r.right()-1, r.top()+2);
}

void QtCurveStyle::drawSliderHandle(QPainter *p, const QRect &r, QStyle::SFlags flags) const
{
    const QColor *fill =
        (flags & (QStyle::Style_Down | QStyle::Style_On | QStyle::Style_Sunken)) ? &button[3] :
        (flags & QStyle::Style_MouseOver) ? &button[1] : &button[6];

    drawLightBevel(p, r, Style_Raised, fill, button);

    if((flags & Style_Horizontal && r.width() >= 14) || r.height() >= 14)
        drawLines(p, r, r.width()<r.height(), 4, 3, button);
}

void QtCurveStyle::drawSliderGroove(QPainter *p, const QRect &r, QStyle::SFlags flags, const QWidget *widget) const
{
    const QSlider *slider=(const QSlider *)widget;
    QRect         groove(r);

    if (flags & Style_HasFocus)
    {
        QRect fr(groove);
        fr.addCoords(-1, -1, 1, 1);
        drawPrimitive(PE_FocusRect, p, fr, QColorGroup());
    }

    if (slider->orientation() == Qt::Horizontal)
    {
        int dh = (groove.height() - 5) / 2;
        groove.addCoords(0, dh, 0, -dh);
    }
    else
    {
        int dw = (groove.width() - 5) / 2;
        groove.addCoords(dw, 0, -dw, 0);
    }
    p->setBrush(gray[2]);
    p->setPen(gray[5]);
    p->drawRect(groove);
    p->setPen(gray[4]);
    p->drawLine(groove.x()+1, groove.y()+1, groove.x()+groove.width()-2, groove.y()+1);
    p->drawLine(groove.x()+1, groove.y()+1, groove.x()+1, groove.y()+groove.height()-2);
}

#include "qtcurve.moc"
